package net.sf.alchim.mojo.yuicompressor;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.util.List;

import org.apache.maven.model.Resource;
import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.project.MavenProject;
import org.codehaus.plexus.util.DirectoryScanner;

import com.yahoo.platform.yui.compressor.CssCompressor;
import com.yahoo.platform.yui.compressor.JavaScriptCompressor;

/**
 * Apply compression on JS and CSS (using YUI Compressor).
 *
 * @goal compress
 * @phase process-resources
 *
 * @author David Bernard
 * @created 2007-08-28
 */
// @SuppressWarnings("unchecked")
public class YuiCompressorMojo extends AbstractMojo {
    private static final String[] EMPTY_STRING_ARRAY = {};

    private static final String[] DEFAULT_INCLUDES = { "**/*.js", "**/*.css" };

    /**
     * Single directory for extra files to include in the WAR.
     *
     * @parameter expression="${basedir}/src/main/webapp"
     */
    private File warSourceDirectory;

    /**
     * The directory where the webapp is built.
     *
     * @parameter expression="${project.build.directory}/${project.build.finalName}"
     */
    private File webappDirectory;

    /**
     * The output directory into which to copy the resources.
     *
     * @parameter expression="${project.build.outputDirectory}"
     */
    private File outputDirectory;

    /**
     * The list of resources we want to transfer.
     *
     * @parameter expression="${project.resources}"
     */
    private List<Resource> resources;

    /**
     * Read the input file using "encoding".
     *
     * @parameter expression="${file.encoding}" default-value="UTF-8"
     */
    private String encoding;

    /**
     * The output filename suffix.
     *
     * @parameter expression="${maven.yuicompressor.suffix}" default-value="-min"
     */
    private String suffix;

    /**
     * [js only] Insert line breaks in output for readability.
     *
     * @parameter expression="${maven.yuicompressor.linebreak}" default-value="false"
     */
    private boolean linebreak;

    /**
     * [js only] Minify only, do not obfuscate.
     *
     * @parameter expression="${maven.yuicompressor.nomunge}" default-value="false"
     */
    private boolean nomunge;

    /**
     * [js only] Display possible errors in the code
     *
     * @parameter expression="${maven.yuicompressor.jswarm}" default-value="true"
     */
    private boolean jswarn;

    /**
     * force the compression of every files,
     * else if compressed file already exists and is younger than source file, nothing is done.
     *
     * @parameter expression="${maven.yuicompressor.force}" default-value="false"
     */
    private boolean force;

    /**
     * show statistics (compression ratio).
     *
     * @parameter expression="${maven.yuicompressor.statistics}" default-value="true"
     */
    private boolean statistics;

    /**
     * define is plugin must stop/fail on warnings.
     *
     * @parameter expression="${maven.yuicompressor.failOnWarning}" default-value="false"
     */
    private boolean failOnWarning;

    /**
     * list of additionnal excludes
     *
     * @parameter
     */
    private List<String> excludes;


    /**
     * @parameter expression="${project}"
     * @required
     */
    protected MavenProject project;

    private long inSizeTotal_;
    private long outSizeTotal_;
    private ErrorReporter4Mojo reporter_;

    @SuppressWarnings("unchecked")
    public void execute() throws MojoExecutionException, MojoFailureException {
        try {
            reporter_ = new ErrorReporter4Mojo(getLog(), jswarn);
            for (Resource resource : resources){
                compressDir(new File( resource.getDirectory() ), outputDirectory, resource.getIncludes(), resource.getExcludes());
            }
            compressDir(warSourceDirectory, webappDirectory, null, null);
            if (statistics) {
                getLog().info(String.format("total input (%db) -> output (%db)[%d%%]", inSizeTotal_, outSizeTotal_, ((outSizeTotal_ * 100)/inSizeTotal_)));
            }
            getLog().info(String.format("nb warnings: %d, nb errors: %d", reporter_.getWarningCnt(), reporter_.getErrorCnt()));
            if (failOnWarning) {
                throw new MojoFailureException("warnings during compression of js and css! (see log)");
            }
        } catch (RuntimeException exc) {
            throw exc;
        } catch (MojoFailureException exc) {
            throw exc;
        } catch (MojoExecutionException exc) {
            throw exc;
        } catch (Exception exc) {
            throw new MojoExecutionException("wrap: " + exc.getMessage(), exc);
        }
    }

    private void compressDir(File srcRoot, File destRoot, List<String> srcIncludes, List<String> srcExcludes) throws Exception {
        if ((srcRoot == null) || ( !srcRoot.exists() )) {
            return;
        }
        if (destRoot == null) {
            throw new MojoFailureException("destination directory for " + srcRoot + " is null");
        }
        DirectoryScanner scanner = new DirectoryScanner();
        scanner.setBasedir(srcRoot);
        if ( (srcIncludes != null) && !srcIncludes.isEmpty() ) {
            scanner.setIncludes( srcIncludes.toArray( EMPTY_STRING_ARRAY ) );
        } else {
            scanner.setIncludes( DEFAULT_INCLUDES );
        }
        if ( (srcExcludes != null) && !srcExcludes.isEmpty() ) {
            scanner.setExcludes( srcExcludes.toArray( EMPTY_STRING_ARRAY ) );
        }
        if ((excludes != null) && !excludes.isEmpty()) {
            scanner.setExcludes( excludes.toArray( EMPTY_STRING_ARRAY ) );
        }
        scanner.addDefaultExcludes();
        scanner.scan();
        for(String name :scanner.getIncludedFiles() ) {
            SourceFile src = new SourceFile(srcRoot, name);
            if (".js".equalsIgnoreCase(src.getExtension()) || ".css".equalsIgnoreCase(src.getExtension())) {
                compressFile(src, destRoot);
            }
        }
    }

    private void compressFile(SourceFile src, File destRoot) throws Exception {
        if (getLog().isDebugEnabled()) {
            getLog().debug("compress file :" + src.toFile()+ " to " + src.toDestFile(destRoot, suffix));
        }
        File inFile = src.toFile();
        File outFile = src.toDestFile(destRoot, suffix);
        if (!force && outFile.exists() && (outFile.lastModified() > src.toFile().lastModified())) {
            if (getLog().isInfoEnabled()) {
                getLog().info("nothing to do, " + outFile + " is younger than original, use 'force' option or clean your target");
            }
            return;
        }
        InputStreamReader in = null;
        OutputStreamWriter out = null;
        try {
            in = new InputStreamReader(new FileInputStream(inFile), encoding);
            if (!outFile.getParentFile().exists() && !outFile.getParentFile().mkdirs()) {
                throw new MojoExecutionException( "Cannot create resource output directory: " + outFile.getParentFile() );
            }
            out = new OutputStreamWriter(new FileOutputStream(outFile), encoding);
            if (".js".equalsIgnoreCase(src.getExtension())) {
                reporter_.setDefaultFileName("..." + inFile.getAbsolutePath().substring(project.getBasedir().getAbsolutePath().length()));
                JavaScriptCompressor compressor = new JavaScriptCompressor(in, reporter_);
                compressor.compress(out, !nomunge, jswarn, linebreak);
            } else if (".css".equalsIgnoreCase(src.getExtension())) {
                CssCompressor compressor = new CssCompressor(in);
                compressor.compress(out);
            }
        } finally {
            close(in);
            close(out);
        }
        if (statistics) {
            inSizeTotal_ += inFile.length();
            outSizeTotal_ += outFile.length();
            getLog().info(String.format("%s (%db) -> %s (%db)[%d%%]", inFile.getName(), inFile.length(), outFile.getName(), outFile.length(), ((outFile.length() * 100)/inFile.length())));
        }
    }

    private void close(InputStreamReader stream) {
        if (stream != null) {
            try {
                stream.close();
            } catch (IOException exc) {
                getLog().warn(exc);
            }
        }
    }

    private void close(OutputStreamWriter stream) {
        if (stream != null) {
            try {
                stream.close();
            } catch (IOException exc) {
                getLog().warn(exc);
            }
        }
    }

}
